use std::{
    collections::HashMap,
    io::{BufRead, BufReader, BufWriter, Write},
    net::{TcpListener, TcpStream},
    sync::Arc,
};

use game::Game;
use stations_list::{FilterTransports, StationsList, ALL_FILTER_TRANSPORTS};

mod game;
mod stations_list;
mod stations_thread;

fn main() {
    let stations = StationsList::new_from("stations_list".into()).unwrap();
    let stations = Arc::new(stations);
    let _thread = stations_thread::spawn(Arc::clone(&stations), 100);
    let listener = TcpListener::bind("0.0.0.0:26021").unwrap();
    let mut threads: Vec<std::thread::JoinHandle<()>> = vec![];
    let max_threads = 8;
    loop {
        if let Ok((mut con, _)) = listener.accept() {
            if let Some(i) = threads
                .iter()
                .enumerate()
                .rev()
                .find(|v| v.1.is_finished())
                .map(|v| v.0)
            {
                threads.swap_remove(i);
            }
            if threads.len() < max_threads {
                let stations = Arc::clone(&stations);
                threads.push(std::thread::spawn(move || {
                    let _ = connection(&mut con, stations);
                    let _ = con.shutdown(std::net::Shutdown::Both);
                }));
            } else {
                let _ = write!(&mut con, "server busy, try again later");
                let _ = con.shutdown(std::net::Shutdown::Both);
            }
        }
    }
}

fn connection(con: &mut TcpStream, stations: Arc<StationsList>) -> std::io::Result<()> {
    let mut w = BufWriter::new(con);
    writeln!(w, "–––=== Bahnhofsreise ===–––")?;
    writeln!(w, "Bekannte Bahnhöfe: {}", stations.station_count())?;
    let filter_transports: FilterTransports = rand::random();
    let mut filter_transports = filter_transports.and_unknown();
    let mut start: Vec<(String, String)> = vec![];
    let mut ziel: Vec<(String, String)> = vec![];
    loop {
        for (i, ft) in ALL_FILTER_TRANSPORTS.iter().enumerate() {
            writeln!(
                w,
                "Kategorie {}: [{}] {}",
                i + 1,
                if *ft == filter_transports { 'x' } else { ' ' },
                ft.explined()
            )?;
        }
        write!(w, " Start: ")?;
        if start.is_empty() {
            writeln!(w, "Zufällig")?;
        } else {
            for (i, v) in start.iter().enumerate() {
                if i != 0 {
                    write!(w, "  /  ")?;
                }
                write!(w, "{}", v.1)?;
            }
            writeln!(w)?;
        }
        write!(w, "  Ziel: ")?;
        if ziel.is_empty() {
            writeln!(w, "Zufällig")?;
        } else {
            for (i, v) in ziel.iter().enumerate() {
                if i != 0 {
                    write!(w, "  /  ")?;
                }
                write!(w, "{}", v.1)?;
            }
            writeln!(w)?;
        }
        writeln!(
            w,
            "Enter zum Starten, oder Kategorie <1-6>, oder Start <Bahnhof>, oder Ziel <Bahnhof>"
        )?;
        w.flush()?;

        let line = BufReader::new(w.get_mut())
            .lines()
            .next()
            .ok_or(std::io::ErrorKind::BrokenPipe)??
            .to_lowercase();
        let line = line.trim();
        match line {
            "kategorie 1" => filter_transports = ALL_FILTER_TRANSPORTS[0],
            "kategorie 2" => filter_transports = ALL_FILTER_TRANSPORTS[1],
            "kategorie 3" => filter_transports = ALL_FILTER_TRANSPORTS[2],
            "kategorie 4" => filter_transports = ALL_FILTER_TRANSPORTS[3],
            "kategorie 5" => filter_transports = ALL_FILTER_TRANSPORTS[4],
            "kategorie 6" => filter_transports = ALL_FILTER_TRANSPORTS[5],
            "" => break,
            line => {
                if line.starts_with("start ") {
                    start = stations.find_stations(line[6..].trim());
                } else if line.starts_with("ziel ") {
                    ziel = stations.find_stations(line[5..].trim());
                }
            }
        }
    }

    if start.is_empty() || ziel.is_empty() {
        writeln!(
            w,
            "Suche nach Bahnhöfen, wo was los ist... (kann etwas dauern)",
        )?;
    } else {
        writeln!(w, "Abfahrt . . .",)?;
    }
    w.flush()?;
    start.truncate(1);
    let start = start.pop();
    ziel.truncate(1);
    let ziel = ziel.pop();
    let custom_ziel = ziel.is_some();
    let mut game = match Game::new(
        filter_transports,
        start.map(|v| v.0),
        ziel.map(|v| v.0),
        Arc::clone(&stations),
    ) {
        Ok(game) => game,
        Err(e) => {
            writeln!(w, "\n[FEHLER] Spielvorbereitung hat nicht geklappt :(\n")?;
            writeln!(w, "{e}")?;
            w.flush()?;
            return Ok(());
        }
    };
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w)?;
    writeln!(w, "→ Dein Startbahnhof: {}.", game.location_name())?;
    writeln!(w, " → Dein Zielbahnhof: {}.", game.target_name())?;
    writeln!(
        w,
        "  → Activity Scores: {} → {}",
        game.location_activity_count(),
        game.target_activity_count()
    )?;
    if custom_ziel && game.target_activity_count() < 3 {
        writeln!(
            w,
            "Geringer Activity Score von {} am Ziel, sicher? [Enter])",
            game.target_activity_count()
        )?;
        w.flush()?;
        BufReader::new(w.get_mut()).lines().next();
    }
    w.flush()?;
    let mut skip_verbindungen = false;
    loop {
        if !skip_verbindungen {
            writeln!(w)?;
            writeln!(w)?;
            writeln!(w, "–– Mögliche Verbindungen ––")?;
            w.flush()?;
            game.update_departures();
            game.location_name();
        }
        let departures = match game.departures_cached() {
            Ok(deps) => deps,
            Err(e) => {
                writeln!(w, "\n[FEHLER] Zugsuche hat nicht geklappt :(\n")?;
                writeln!(w, "{}", e)?;
                w.flush()?;
                return Ok(());
            }
        };
        let mut possible_locations = HashMap::<String, &str>::new();
        for zug in &departures.entries {
            for linie in zug {
                if linie.canceled {
                    if !skip_verbindungen {
                        writeln!(
                            w,
                            "[fällt aus] {} | {} → {}",
                            linie.line_name, linie.stop_place.name, linie.destination.name
                        )?;
                    }
                } else {
                    if !skip_verbindungen {
                        writeln!(
                            w,
                            "{}  |  {}  →  {}",
                            linie.line_name, linie.stop_place.name, linie.destination.name
                        )?;
                    }
                    possible_locations.insert(
                        linie.destination.name.trim().to_lowercase(),
                        &linie.destination.slug,
                    );
                    if !linie.via_stops.is_empty() {
                        if !skip_verbindungen {
                            write!(w, "  via")?;
                        }
                        let mut width = 3;
                        for (istop, stop) in linie.via_stops.iter().enumerate() {
                            if !skip_verbindungen {
                                if istop > 0 {
                                    write!(w, ",")?;
                                    width += 1;
                                }
                                if width > 0 && width + stop.name.len() > 70 {
                                    writeln!(w)?;
                                    write!(w, "  ")?;
                                    width = 0;
                                }
                                if istop == 0 {
                                    width += 1 + stop.name.len();
                                    write!(w, " {}", stop.name)?;
                                } else {
                                    width += 4 + stop.name.len();
                                    write!(w, "    {}", stop.name)?;
                                }
                            }
                            possible_locations.insert(stop.name.trim().to_lowercase(), &stop.slug);
                        }
                        if !skip_verbindungen {
                            writeln!(w)?;
                        }
                    }
                }
                if !skip_verbindungen {
                    writeln!(w)?;
                }
            }
            if !skip_verbindungen {
                writeln!(w)?;
            }
        }
        if !skip_verbindungen {
            writeln!(w)?;
            writeln!(w)?;
            writeln!(w, "→ Aktueller Bahnhof: {}.", game.location_name_immut())?;
            writeln!(w, " → Dein Zielbahnhof: {}.", game.target_name_immut())?;
        }
        skip_verbindungen = false;
        let mut r_errs = 0;
        loop {
            writeln!(w, "Wohin geht die Reise? (`?` für Hilfe)")?;
            w.flush()?;
            match BufReader::new(w.get_mut()).lines().next() {
                None => return Ok(()),
                Some(line) => {
                    let line = line?;
                    let line = line.trim();
                    if line == "?" {
                        writeln!(w)?;
                        writeln!(w, "–– Erklärung ––")?;
                        writeln!(
                            w,
                            "Bahnreise ist ein Spiel basierend auf Live-Daten von `bahnhof.de`."
                        )?;
                        writeln!(
                            w,
                            "Ziel ist es, vom Startbahnhof aus den Zielbahnhof zu erreichen."
                        )?;
                        writeln!(
                            w,
                            "Das Spiel zeigt eine Liste von Verbindungen an, die an dem Bahnhof"
                        )?;
                        writeln!(
                            w,
                            "abfahren, an dem du gerade bist. Um irgendwohin zu fahren, tippe"
                        )?;
                        writeln!(w, "den Namen eines Bahnhofs eine und bestätige mit Enter.")?;
                        writeln!(
                            w,
                            "Das wiederholst du, bis du am Zielbahnhof angekommen bist."
                        )?;
                        writeln!(
                            w,
                            "Wenn du nicht weißt, wo der Zielbahnhof ist, kannst du dir"
                        )?;
                        writeln!(
                            w,
                            "mit `??` einen Tipp anzeigen lassen, der dir verrät, von wo aus"
                        )?;
                        writeln!(
                            w,
                            "du dein Ziel erreichen kannst. Sonst hilft nur noch eine Landkarte."
                        )?;
                    } else if line == "??" {
                        let target_name = game.target_name_immut().to_owned();
                        match game.get_hint() {
                            Ok(hint) => {
                                let mut arrivals: Vec<(&'_ str, Vec<&'_ str>)> = vec![];
                                for arrival in hint.entries.iter().map(|v| v.iter()).flatten() {
                                    if let Some((_, lines)) = arrivals
                                        .iter_mut()
                                        .find(|v| v.0 == arrival.origin.name.as_str())
                                    {
                                        if !arrival.canceled {
                                            if !lines.contains(&arrival.line_name.as_str()) {
                                                lines.push(arrival.line_name.as_str());
                                            }
                                        }
                                    } else {
                                        arrivals.push((
                                            arrival.origin.name.as_str(),
                                            if !arrival.canceled {
                                                vec![arrival.line_name.as_str()]
                                            } else {
                                                vec![]
                                            },
                                        ));
                                    }
                                }
                                writeln!(w, "\n\n[TIPP] Züge, die in {target_name} ankommen:")?;
                                for (origin, lines) in arrivals {
                                    writeln!(
                                        w,
                                        "- {origin} ({})",
                                        if lines.is_empty() {
                                            format!("all connections cancelled")
                                        } else {
                                            lines
                                                .into_iter()
                                                .enumerate()
                                                .map(|(i, v)| {
                                                    format!("{}{v}", if i == 0 { "" } else { ", " })
                                                })
                                                .collect()
                                        }
                                    )?;
                                }
                            }
                            Err(e) => {
                                writeln!(w, "ups, das hat nicht geklappt... ({e})")?;
                            }
                        }
                        skip_verbindungen = true;
                        break;
                    } else if line.is_empty() {
                        r_errs += 1;
                        if r_errs > 25 {
                            return Ok(());
                        }
                    } else if let Some(new) = possible_locations
                        .get(line.to_lowercase().as_str())
                        .map(|v| (*v).to_owned())
                        .or_else(|| {
                            stations
                                .find_stations(line)
                                .into_iter()
                                .filter_map(|(id, _)| {
                                    possible_locations
                                        .values()
                                        .find(|v| **v == id.as_str())
                                        .map(|_| id.clone())
                                })
                                .next()
                        })
                    {
                        if new.trim().is_empty() {
                            writeln!(
                                w,
                                "Diesen Halt gibt es scheinbar, aber er hat keine Bahnhofs-Seite,"
                            )?;
                            writeln!(w, "d.h. es können leider auch keine Infos/Abfahrten erfragt werden :(")?;
                            w.flush()?;
                        } else {
                            match game.go_to_check(new.as_str()) {
                                Ok(0) if game.target() != new.as_str() => {
                                    writeln!(w, "sieht dort recht leer aus... (this is softlock protection btw)")?;
                                    w.flush()?;
                                }
                                Ok(act) => {
                                    let mut go_to_station = true;
                                    if act <= 3 && game.target() != new.as_str() {
                                        writeln!(
                                            w,
                                            "Geringer Activity Score von {act}, sicher? [ja/nein])"
                                        )?;
                                        w.flush()?;
                                        go_to_station = false;
                                        if let Some(Ok(v)) =
                                            BufReader::new(w.get_mut()).lines().next()
                                        {
                                            if v.as_str() == "ja" {
                                                go_to_station = true;
                                            }
                                        }
                                    }
                                    if go_to_station {
                                        let location = new;
                                        if game.go_to_unchecked(location) {
                                            writeln!(w)?;
                                            writeln!(w)?;
                                            writeln!(w, "yay, du bist da :D")?;
                                            let ilen = game.history().len().to_string().len();
                                            for (i, (id, name)) in game.history().iter().enumerate()
                                            {
                                                let i = (i + 1).to_string();
                                                writeln!(
                                                    w,
                                                    "- {}{i}. {}",
                                                    " ".repeat(ilen.saturating_sub(i.len())),
                                                    name.as_ref()
                                                        .map(|v| v.as_str())
                                                        .unwrap_or(id.as_str())
                                                )?;
                                            }
                                            writeln!(
                                                w,
                                                "- {} → {}",
                                                " ".repeat(ilen.saturating_sub(1)),
                                                game.location_name()
                                            )?;
                                            w.flush()?;
                                            return Ok(());
                                        } else {
                                            break;
                                        }
                                    }
                                }
                                Err(e) => {
                                    writeln!(w, "hm, vielleicht lieber nicht dorthin? ({e})")?;
                                    w.flush()?;
                                }
                            }
                        }
                    } else {
                        writeln!(w, "hm, kenne ich nicht... (typo?)")?;
                        r_errs += 1;
                        if r_errs > 25 {
                            return Ok(());
                        }
                    }
                }
            }
        }
    }
}
